// **************************************************************************
// * This file is part of the zenXML project. It is distributed under the   *
// * Boost Software License, Version 1.0. See accompanying file             *
// * LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt.       *
// * Copyright (C) 2011 ZenJu (zhnmju123 AT gmx.de)                         *
// **************************************************************************

#ifndef ZEN_XML_CONVERT_HEADER_018727409908342709743
#define ZEN_XML_CONVERT_HEADER_018727409908342709743

#include "string_utf8.h"
#include "string_tools.h"
//#include <loki/TypeTraits.h> -> include implicitly via <string_tools.h>! -> specialized __int64
#include "zenxml_dom.h"

namespace zen
{
/**
\file
\brief Handle conversion of arbitrary types to and from XML elements.

It is \b not required to call these functions directly. They are implicitly used by zen::XmlElement::getValue(),
zen::XmlElement::setValue(), zen::XmlElement::getAttribute() and zen::XmlElement::setAttribute().
\n\n
By default the following conversions to user types are automatically supported:
	- strings - std::string, std::wstring, char*, wchar_t*, char, wchar_t, ect..., all STL-compatible-string-classes
	- numbers - int, double, float, bool, long, ect..., all built-in numbers
	- STL containers - std::map, std::set, std::vector, std::list, ect..., all STL-compatible-containers
	- std::pair

You can add support for additional types via template specialization. \n\n
Specialize zen::readValue() and zen::writeValue() to enable conversion from structured user types to XML elements.
Specialize zen::readText() and zen::writeText() to enable conversion from string-convertible user types to std::string.
Latter is generally more useful as it does not only allow converting from XML elements to user data, but also from and to XML attributes.
\n\n
<b> Example: </b> type "bool"
\code
namespace zen
{
template <> inline
void writeText(const bool& value, std::string& output)
{
    output = value ? "true" : "false";
}

template <> inline
bool readText(const std::string& input, bool& value)
{
    std::string tmp = input;
    zen::trim(tmp);
    if (tmp == "true")
        value = true;
    else if (tmp == "false")
        value = false;
    else
        return false;
    return true;
}
}
\endcode
*/

///Convert XML element to user data
/**
  \param input The input XML element.
  \param value Conversion target value.
  \return "true" if value was read successfully.
  \sa zenxml_cvrt.h
*/
template <class T> bool readValue(const XmlElement& input, T& value);
///Convert user data into an XML element
/**
  \param value The value to be converted.
  \param output The output XML element.
  \sa zenxml_cvrt.h
*/
template <class T> void writeValue(const T& value, XmlElement& output);


///Convert text to user data - used by XML elements and attributes
/**
  \param input Input text.
  \param value Conversion target value.
  \return "true" if value was read successfully.
  \sa zenxml_cvrt.h
*/
template <class T> bool readText(const std::string& input, T& value);
///Convert user data into text - used by XML elements and attributes
/**
  \param value The value to be converted.
  \param output Output text.
  \sa zenxml_cvrt.h
*/
template <class T> void writeText(const T& value, std::string& output);


/* Different classes of data types:

---------------------------
| structured              |  readValue/writeValue - e.g. string-convertible types, STL containers, std::pair, structured user types
| ----------------------  |
| | string-convertible |  |  readText/writeText   - e.g. string-like types, all built-in arithmetic numbers, bool
| | ---------------    |  |
| | | string-like |    |  |  utf8CvrtTo           - e.g. std::string, wchar_t*, char[], wchar_t, wxString, MyStringClass, ...
| | ---------------    |  |
| ----------------------  |
---------------------------
*/






















//------------------------------ implementation -------------------------------------
namespace implementation
{
template <typename T>
class HasStlContainerMemberTypes
{
    typedef char Yes[1];
    typedef char No [2];

    //detect presence of member type names
    template <typename U> class HelperTp {};

    template <class U> static Yes& hasMemberValueType(HelperTp<typename U::value_type>*);
    template <class U> static  No& hasMemberValueType(...);

    template <class U> static Yes& hasMemberIterator(HelperTp<typename U::iterator>*);
    template <class U> static  No& hasMemberIterator(...);

    template <class U> static Yes& hasMemberConstIterator(HelperTp<typename U::const_iterator>*);
    template <class U> static  No& hasMemberConstIterator(...);

public:
    enum { Result = sizeof(hasMemberValueType    <T>(NULL)) == sizeof(Yes) &&
                    sizeof(hasMemberIterator     <T>(NULL)) == sizeof(Yes) &&
                    sizeof(hasMemberConstIterator<T>(NULL)) == sizeof(Yes)
         };
};


template <typename T, bool isClassType>
class HasStlContainerMemberFun
{
public:
    enum { Result = false};
};

template <typename T>
class HasStlContainerMemberFun<T, true> //T must be class-type!
{
    typedef char Yes[1];
    typedef char No [2];

    //detect presence of member variables (within T or one of it's base classes)
    template <typename U, U t> class HelperFn {};

    struct Fallback
    {
        int begin;
        int end;
        int insert;
    };

    template <class U>
    struct Helper2 : public U, public Fallback {}; //U must be a class-type!

    //we don't know the exact declaration of the member attribute (may be in base class), but we know what NOT to expect:
    template <class U> static  No& hasMemberBegin(HelperFn<int Fallback::*, &Helper2<U>::begin>*);
    template <class U> static Yes& hasMemberBegin(...);

    template <class U> static  No& hasMemberEnd(HelperFn<int Fallback::*, &Helper2<U>::end>*);
    template <class U> static Yes& hasMemberEnd(...);

    template <class U> static  No& hasMemberInsert(HelperFn<int Fallback::*, &Helper2<U>::insert>*);
    template <class U> static Yes& hasMemberInsert(...);

public:
    enum { Result = sizeof(hasMemberBegin <T>(NULL)) == sizeof(Yes) &&
                    sizeof(hasMemberEnd   <T>(NULL)) == sizeof(Yes) &&
                    sizeof(hasMemberInsert<T>(NULL)) == sizeof(Yes)
         };
};
}

template <typename T>
class IsStlContainer
{
public:
    enum { Result = implementation::HasStlContainerMemberFun<T, implementation::HasStlContainerMemberTypes<T>::Result>::Result };
};


namespace implementation
{
template <typename T>
class HasStlPairMemberTypes
{
    typedef char Yes[1];
    typedef char No [2];

    //detect presence of member type names
    template <typename U> class HelperTp {};

    template <class U> static Yes& hasMemberFirstType(HelperTp<typename U::first_type>*);
    template <class U> static  No& hasMemberFirstType(...);

    template <class U> static Yes& hasMemberSecondType(HelperTp<typename U::second_type>*);
    template <class U> static  No& hasMemberSecondType(...);

public:
    enum { Result = sizeof(hasMemberFirstType <T>(NULL)) == sizeof(Yes) &&
                    sizeof(hasMemberSecondType<T>(NULL)) == sizeof(Yes)
         };
};


template <typename T, bool isClassType>
class HasStlPairMemberVar
{
public:
    enum { Result = false};
};


template <typename T>
class HasStlPairMemberVar<T, true> //T must be class-type!
{
    typedef char Yes[1];
    typedef char No [2];

    //detect presence of member variables (within T or one of it's base classes)
    template <typename U, U t> class HelperFn {};

    struct Fallback
    {
        int first;
        int second;
    };

    template <class U>
    struct Helper2 : public U, public Fallback {}; //U must be a class-type!

    //we don't know the exact declaration of the member attribute (may be in base class), but we know what NOT to expect:
    template <class U> static No& hasMemberFirst(HelperFn<int Fallback::*, &Helper2<U>::first>*);
    template <class U> static Yes& hasMemberFirst(...);

    template <class U> static No& hasMemberSecond(HelperFn<int Fallback::*, &Helper2<U>::second>*);
    template <class U> static Yes& hasMemberSecond(...);

public:
    enum { Result = sizeof(hasMemberFirst <T>(NULL)) == sizeof(Yes) &&
                    sizeof(hasMemberSecond<T>(NULL)) == sizeof(Yes)
         };
};
}


template <typename T>
class IsStlPair
{
public:
    enum { Result = implementation::HasStlPairMemberVar<T, implementation::HasStlPairMemberTypes<T>::Result>::Result };
};
//######################################################################################

//Conversion from arbitrary types to text (for use with XML elements and attributes)
enum TextType
{
    TEXT_TYPE_BOOL,
    TEXT_TYPE_NUMBER,
    TEXT_TYPE_STRING,
    TEXT_TYPE_OTHER,
};


template <class T>
struct GetTextType
{
    static const TextType value = Loki::IsSameType<T, bool>::value ? TEXT_TYPE_BOOL :
                                  StringTraits<T>::isStringLike    ? TEXT_TYPE_STRING : //string before number to correctly handle char/wchar_t
                                  Loki::TypeTraits<T>::isArith     ? TEXT_TYPE_NUMBER : //
                                  TEXT_TYPE_OTHER;
};


template <class T, TextType type>
struct ConvertText;
/* -> expected interface
{
	void writeText(const T& value, std::string& output) const;
	bool readText(const std::string& input, T& value) const;
};
*/

//partial specialization: type bool
template <class T>
struct ConvertText<T, TEXT_TYPE_BOOL>
{
    void writeText(bool value, std::string& output) const
    {
        output = value ? "true" : "false";
    }
    bool readText(const std::string& input, bool& value) const
    {
        std::string tmp = input;
        zen::trim(tmp);
        if (tmp == "true")
            value = true;
        else if (tmp == "false")
            value = false;
        else
            return false;
        return true;
    }
};

//partial specialization: handle conversion for all built-in arithmetic types!
template <class T>
struct ConvertText<T, TEXT_TYPE_NUMBER>
{
    void writeText(const T& value, std::string& output) const
    {
        output = toString<std::string>(value);
    }
    bool readText(const std::string& input, T& value) const
    {
        value = toNumber<T>(input);
        return true;
    }
};

//partial specialization: handle conversion for all string-like types!
template <class T>
struct ConvertText<T, TEXT_TYPE_STRING>
{
    void writeText(const T& value, std::string& output) const
    {
        output = utf8CvrtTo<std::string>(value);
    }
    bool readText(const std::string& input, T& value) const
    {
        value = utf8CvrtTo<T>(input);
        return true;
    }
};


//partial specialization: unknown type
template <class T>
struct ConvertText<T, TEXT_TYPE_OTHER>
{
    //###########################################################################################################################################
    assert_static(sizeof(T) == -1);
    /*
        ATTENTION: The data type T is yet unknown to the zenXML framework!

        Please provide a specialization for T of the following two functions in order to handle conversions to XML elements and attributes

    	template <> void zen::writeText(const T& value, std::string& output)
        template <> bool zen::readText(const std::string& input, T& value)

    	If T is structured and cannot be converted to a text representation specialize these two functions to allow at least for conversions to XML elements:

    	template <> void zen::writeValue(const T& value, XmlElement& output)
        template <> bool zen::readValue(const XmlElement& input, T& value)
    */
    //###########################################################################################################################################
};


template <class T> inline
void writeText(const T& value, std::string& output)
{
    ConvertText<T, GetTextType<T>::value>().writeText(value, output);
}


template <class T> inline
bool readText(const std::string& input, T& value)
{
    return ConvertText<T, GetTextType<T>::value>().readText(input, value);
}
//######################################################################################

//Conversion from arbitrary types to an XML element
enum ValueType
{
    VALUE_TYPE_STL_CONTAINER,
    VALUE_TYPE_STL_PAIR,
    VALUE_TYPE_OTHER,
};

template <class T>
struct GetValueType
{
    static const ValueType value = GetTextType<T>::value != TEXT_TYPE_OTHER ? VALUE_TYPE_OTHER : //some string classes are also STL containers, so check this first
                                   IsStlContainer<T>::Result                ? VALUE_TYPE_STL_CONTAINER :
                                   IsStlPair<T>::Result                     ? VALUE_TYPE_STL_PAIR :
                                   VALUE_TYPE_OTHER;
};


template <class T, ValueType type>
struct ConvertElement;
/* -> expected interface
{
	void writeValue(const T& value, XmlElement& output) const;
	bool readValue(const XmlElement& input, T& value) const;
};
*/


//partial specialization: handle conversion for all STL-container types!
template <class T>
struct ConvertElement<T, VALUE_TYPE_STL_CONTAINER>
{
    void writeValue(const T& value, XmlElement& output) const
    {
        std::for_each(value.begin(), value.end(),
                      [&](const typename T::value_type& childVal)
        {
            XmlElement& newChild = output.addChild("Item");
            zen::writeValue(childVal, newChild);
        });
    }
    bool readValue(const XmlElement& input, T& value) const
    {
        bool success = true;
        value.clear();

        auto iterPair = input.getChildren("Item");
        for (auto iter = iterPair.first; iter != iterPair.second; ++iter)
        {
            typename T::value_type childVal; //MSVC 2010 bug: cannot put this into a lambda body
            if (zen::readValue(*iter, childVal))
                value.insert(value.end(), childVal);
            else
                success = false;
        }
        return success;
    }
};


//partial specialization: handle conversion for std::pair
template <class T>
struct ConvertElement<T, VALUE_TYPE_STL_PAIR>
{
    void writeValue(const T& value, XmlElement& output) const
    {
        XmlElement& child1 = output.addChild("1st");
        zen::writeValue(value.first, child1);

        XmlElement& child2 = output.addChild("2nd");
        zen::writeValue(value.second, child2);
    }
    bool readValue(const XmlElement& input, T& value) const
    {
        bool success = true;
        const XmlElement* child1 = input.getChild("1st");
        if (!child1 || !zen::readValue(*child1, value.first))
            success = false;

        const XmlElement* child2 = input.getChild("2nd");
        if (!child2 || !zen::readValue(*child2, value.second))
            success = false;

        return success;
    }
};


//partial specialization: not a pure structured type, try text conversion (thereby respect user specializations of writeText()/readText())
template <class T>
struct ConvertElement<T, VALUE_TYPE_OTHER>
{
    void writeValue(const T& value, XmlElement& output) const
    {
        std::string tmp;
        writeText(value, tmp);
        output.setValue(tmp);
    }
    bool readValue(const XmlElement& input, T& value) const
    {
        std::string rawStr;
        input.getValue(rawStr);
        return readText(rawStr, value);
    }
};


template <class T> inline
void writeValue(const T& value, XmlElement& output)
{
    ConvertElement<T, GetValueType<T>::value>().writeValue(value, output);
}


template <class T> inline
bool readValue(const XmlElement& input, T& value)
{
    return ConvertElement<T, GetValueType<T>::value>().readValue(input, value);
}
}

#endif //ZEN_XML_CONVERT_HEADER_018727409908342709743
