// **************************************************************************
// * This file is part of the zenXML project. It is distributed under the   *
// * Boost Software License, Version 1.0. See accompanying file             *
// * LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt.       *
// * Copyright (C) 2011 ZenJu (zhnmju123 AT gmx.de)                         *
// **************************************************************************

#ifndef ZEN_XML_IO_HEADER_8917640501480763248343343
#define ZEN_XML_IO_HEADER_8917640501480763248343343

#include "string_tools.h"
#include "loki/ScopeGuard.h"
#include "zenxml_error.h"
#include "string_utf8.h"
#include <cerrno>

namespace zen
{
/**
\file
\brief Save and load byte streams from files
*/

#if !defined(ZEN_PLATFORM_WINDOWS) && !defined(ZEN_PLATFORM_OTHER)
#error Please specify your platform: #define ZEN_PLATFORM_WINDOWS or ZEN_PLATFORM_OTHER
#endif

///Exception thrown due to failed file I/O
struct XmlFileError : public XmlError
{
    typedef int ErrorCode;

    XmlFileError(ErrorCode ec) : lastError(ec) {}
    ///Native error code: errno
    ErrorCode lastError;
};

#ifdef ZEN_PLATFORM_WINDOWS
template <class String> inline
std::wstring toWideString(const String& str, char) { return utf8ToWide<std::wstring>(str); } //convert UTF8 to wide character string

template <class String> inline
const String& toWideString(const String& str, wchar_t) { return str; } //directly process string without conversion
#endif

///Save byte stream to a file
/**
\tparam String Arbitrary string-like type: e.g. std::string, wchar_t*, char[], wchar_t, wxString, MyStringClass, ...
\param stream Input byte stream
\param filename Output file name
\throw XmlFileError
*/
template <class String>
void saveStream(const std::string& stream, const String& filename) //throw XmlFileError
{
#ifdef ZEN_PLATFORM_WINDOWS
    FILE* handle = ::_wfopen(strBegin(toWideString(filename, typename StringTraits<String>::CharType())), L"wb");
#else
    FILE* handle = ::fopen(toStdString(filename).c_str(), "w");
#endif
    if (handle == NULL)
        throw XmlFileError(errno);

    Loki::ScopeGuard dummy = Loki::MakeGuard(::fclose, handle);
    (void)dummy;

    const size_t bytesWritten = ::fwrite(stream.c_str(), 1, stream.size(), handle);
    if (::ferror(handle) != 0)
        throw XmlFileError(errno);

    assert(bytesWritten == stream.size());
}


///Load byte stream from a file
/**
\tparam String Arbitrary string-like type: e.g. std::string, wchar_t*, char[], wchar_t, wxString, MyStringClass, ...
\param filename Input file name
\return Output byte stream
\throw XmlFileError
*/
template <class String>
std::string loadStream(const String& filename) //throw XmlFileError
{
#ifdef ZEN_PLATFORM_WINDOWS
    FILE* handle = ::_wfopen(strBegin(toWideString(filename, typename StringTraits<String>::CharType())), L"rb");
#else
    FILE* handle = ::fopen(toStdString(filename).c_str(), "r");
#endif
    if (handle == NULL)
        throw XmlFileError(errno);

    Loki::ScopeGuard dummy = Loki::MakeGuard(::fclose, handle);
    (void)dummy;

    std::string stream;
    do
    {
        const size_t BLOCK_SIZE = 10 * 1024;
        char buffer[BLOCK_SIZE];

        const size_t bytesRead = ::fread(buffer, 1, BLOCK_SIZE, handle);
        if (::ferror(handle))
            throw XmlFileError(errno);

        assert(bytesRead <= BLOCK_SIZE);

        stream.append(buffer, bytesRead);
    }
    while (!::feof(handle));

    return stream;
}
}

#endif //ZEN_XML_IO_HEADER_8917640501480763248343343
